package org.netbeans.modules.cnd.dwarfdump.dwarf;

import java.io.IOException;

import org.netbeans.modules.cnd.dwarfdump.dwarfconsts.OP;
import org.netbeans.modules.cnd.dwarfdump.reader.ByteStreamReader;

import com.xmos.libs.core.utils.RandomAccessByteArrayDataInput;

/**
 * Static methods for reading DwarfExpressions 
 * @author Richard Osborne
 *
 */
public class DwarfExpressions {

    private static DwarfExpressionOperation readExpressionOperation(ByteStreamReader dataReader) throws IOException {
		// TODO error handling
		// Read opcode
		OP opcode = OP.get(dataReader.readUnsignedByte());
		DwarfExpressionOperation operation = new DwarfExpressionOperation(opcode);
		// Read operands
		switch (opcode) {
		// One operand opcodes
		case DW_OP_deref_size:
		case DW_OP_xderef_size:
			operation.addOperand(Byte.valueOf(dataReader.readByte()));
			break;
		case DW_OP_skip:
		case DW_OP_bra:
		case DW_OP_call2:
			operation.addOperand(Short.valueOf(dataReader.readShort()));
			break;
		case DW_OP_call4:
			operation.addOperand(Integer.valueOf(dataReader.readShort()));
			break;
		case DW_OP_call_ref:
			operation.addOperand(Long.valueOf(dataReader.read3264()));
			break;
		case DW_OP_addr:
			operation.addOperand(Long.valueOf(dataReader.readAddress()));
			break;
		case DW_OP_plus_uconst:
		case DW_OP_regx:
		case DW_OP_piece:
			operation.addOperand(Integer.valueOf(dataReader.readUnsignedLEB128()));
			break;
		case DW_OP_fbreg:
			operation.addOperand(Integer.valueOf(dataReader.readSignedLEB128()));
			break;
		// Two operand opcodes
		case DW_OP_bregx:
			operation.addOperand(Integer.valueOf(dataReader.readUnsignedLEB128()));
			operation.addOperand(Integer.valueOf(dataReader.readSignedLEB128()));
			break;
		case DW_OP_bit_piece:
			operation.addOperand(Integer.valueOf(dataReader.readSignedLEB128()));
			operation.addOperand(Integer.valueOf(dataReader.readSignedLEB128()));
		default:
			if (opcode.value() >= OP.DW_OP_breg0.value() &&
					opcode.value() <= OP.DW_OP_breg31.value()) {
    			operation.addOperand(Integer.valueOf(dataReader.readSignedLEB128()));
			}
		}
		return operation;
    }
    
    public static DwarfExpression readExpression(ByteStreamReader reader) throws IOException {
    	int size = reader.readShort();
    	DwarfExpression expression = readExpression(reader, size);
    	return expression;
    }
    
    private static DwarfExpression readExpression(ByteStreamReader dataReader, int size) throws IOException {
    	DwarfExpression expression = new DwarfExpression();
    	long end = dataReader.getFilePointer() + size;
    	while (dataReader.getFilePointer() < end) {
    		expression.addOperation(readExpressionOperation(dataReader));
    	}
    	return expression;
    }
    
    public static DwarfExpression readExpression(ByteStreamReader reader, byte[] data) {
    	ByteStreamReader dataReader = new ByteStreamReader(reader.getFileName(),
    													   new RandomAccessByteArrayDataInput(data));
    	dataReader.setAddressSize(reader.getAddressSize());
    	dataReader.setDataEncoding(reader.getDataEncoding());
    	dataReader.setFileClass(reader.getFileClass());

    	DwarfExpression expression;
    	try {
	    	expression = readExpression(dataReader, data.length);
    	} catch (IOException e) {
    		// Shouldn't get here
    		throw new RuntimeException(e);
    	}
    	return expression;
    }
}
